from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pdm.backend.config import Config

if TYPE_CHECKING:
    from typing import Protocol
else:
    Protocol = object


@dataclasses.dataclass()
class Context:
    """The context object for the build hook,
    which contains useful information about the building process.
    Custom hooks can also change the values of attributes or
    assign arbitrary attributes to this object.

    Attributes:
        root: The project root directory
        config: The parsed pyproject.toml as a Config object
        target: The target to build, one of "sdist", "wheel", "editable"
        build_dir: The build directory for storing files generated during the build
        dist_dir: The directory to store the built artifacts
        config_settings: The config settings passed to the hook
        kwargs: The extra args passed to the build method
    """

    root: Path
    config: Config
    target: str
    build_dir: Path
    dist_dir: Path
    config_settings: dict[str, str]
    kwargs: dict[str, Any]

    def ensure_build_dir(self) -> None:
        """Create if the build dir doesn't exist"""
        if not self.build_dir.exists():
            self.build_dir.mkdir(mode=0o700, parents=True)


class BuildHookInterface(Protocol):
    """The interface definition for build hooks.

    Custom hooks can implement part of the methods to provide corresponding abilities.
    """

    def is_enabled(self, context: Context) -> bool:
        """Return True if the hook is enabled for the current build and context

        Parameters:
            context: The context for this build
        """
        ...

    def pdm_build_clean(self, context: Context) -> None:
        """An optional clean step which will be called before the build starts

        Parameters:
            context: The context for this build
        """
        ...

    def pdm_build_initialize(self, context: Context) -> None:
        """This hook will be called before the build starts,
        any updates to the context object will be seen by the following processes.
        It is recommended to modify the metadata in this hook.

        Parameters:
            context: The context for this build
        """
        ...

    def pdm_build_update_files(self, context: Context, files: dict[str, Path]) -> None:
        """Passed in the current file mapping of {relpath: path}
        for hooks to update.

        Parameters:
            context: The context for this build
            files: The file mapping to be included in the build artifact, where the key
                is the relpath inside the artifact(wheel or tarball) and the value is
                the local path to the file.
        """
        ...

    def pdm_build_finalize(self, context: Context, artifact: Path) -> None:
        """This hook will be called after the build is done,
        the artifact is the path to the built artifact.

        Parameters:
            context: The context for this build
            artifact: The path to the built artifact
        """
        ...

    def pdm_build_update_setup_kwargs(
        self, context: Context, kwargs: dict[str, Any]
    ) -> None:
        """Passed in the setup kwargs for hooks to update.

        Parameters:
            context: The context for this build
            kwargs: The arguments to be passed to the setup() function
        """
        ...



from pdm.backend.utils import expand_vars


def test_expand_vars(monkeypatch):
    monkeypatch.setenv("FOO", "foo=a")
    monkeypatch.setenv("BAR", "bar")
    root = "/abc/def"

    line = "file:///${PROJECT_ROOT}/${FOO}:${BAR}:${BAZ}"
    assert expand_vars(line, root) == "file:///abc/def/foo%3Da:bar:${BAZ}"



from ..pre_compile import cli
from . import _deprecated


def run(*args, **kwargs):
    return _deprecated(run, cli.run)(*args, **kwargs)


def main(*args, **kwargs):
    return _deprecated(run, cli.main)(*args, **kwargs)



from textwrap import dedent

from ._vendor.fastjsonschema import (
    JsonSchemaDefinitionException,
    JsonSchemaException,
    JsonSchemaValueException,
)
from .error_reporting import ValidationError


class InvalidSchemaVersion(JsonSchemaDefinitionException):
    """\
    All schemas used in the validator should be specified using the same version \
    as the toplevel schema ({version!r}).

    Schema for {name!r} has version {given!r}.
    """

    def __init__(self, name: str, given_version: str, required_version: str):
        msg = dedent(self.__doc__ or "").strip()
        msg = msg.format(name=name, version=required_version, given=given_version)
        super().__init__(msg)


class SchemaMissingId(JsonSchemaDefinitionException):
    """\
    All schemas used in the validator MUST define a unique toplevel `"$id"`.
    No `"$id"` was found for schema associated with {reference!r}.
    """

    def __init__(self, reference: str):
        msg = dedent(self.__doc__ or "").strip()
        super().__init__(msg.format(reference=reference))


class SchemaWithDuplicatedId(JsonSchemaDefinitionException):
    """\
    All schemas used in the validator MUST define a unique toplevel `"$id"`.
    `$id = {schema_id!r}` was found at least twice.
    """

    def __init__(self, schema_id: str):
        msg = dedent(self.__doc__ or "").strip()
        super().__init__(msg.format(schema_id=schema_id))


__all__ = [
    "InvalidSchemaVersion",
    "JsonSchemaDefinitionException",
    "JsonSchemaException",
    "JsonSchemaValueException",
    "SchemaMissingId",
    "SchemaWithDuplicatedId",
    "ValidationError",
]



from __future__ import annotations

import os
import tarfile
import tempfile
from copy import copy
from pathlib import Path
from typing import Iterable

from pdm.backend.base import Builder
from pdm.backend.hooks import Context
from pdm.backend.structures import FileMap


def normalize_file_permissions(st_mode: int) -> int:
    """
    Normalizes the permission bits in the st_mode field from stat to 644/755

    Popular VCSs only track whether a file is executable or not. The exact
    permissions can vary on systems with different umasks. Normalising
    to 644 (non executable) or 755 (executable) makes builds more reproducible.
    """
    # Set 644 permissions, leaving higher bits of st_mode unchanged
    new_mode = (st_mode | 0o644) & ~0o133
    if st_mode & 0o100:
        new_mode |= 0o111  # Executable: 644 -> 755

    return new_mode


def clean_tarinfo(tar_info: tarfile.TarInfo) -> tarfile.TarInfo:
    """
    Clean metadata from a TarInfo object to make it more reproducible.

        - Set uid & gid to 0
        - Set uname and gname to ""
        - Normalise permissions to 644 or 755
        - Set mtime if not None
    """
    ti = copy(tar_info)
    ti.uid = 0
    ti.gid = 0
    ti.uname = ""
    ti.gname = ""
    ti.mode = normalize_file_permissions(ti.mode)

    return ti


class SdistBuilder(Builder):
    """This build should be performed for PDM project only."""

    target = "sdist"

    def initialize(self, context: Context) -> None:
        super().initialize(context)
        # Save the config to build_dir/pyprojec.toml so that any
        # modification to the pyproject.toml will be saved in the sdist.
        context.ensure_build_dir()
        context.config.write_to(context.build_dir / "pyproject.toml")

    def _collect_files(self, context: Context, root: Path) -> FileMap:
        files = super()._collect_files(context, root)
        local_hook = self.config.build_config.custom_hook

        additional_files: Iterable[str] = filter(
            None,
            (
                local_hook,
                self.config.metadata.readme_file,
                "pyproject.toml",
                *self.find_license_files(),
            ),
        )

        for file in additional_files:
            if root.joinpath(file).exists():
                files[file] = root / file

        return files

    def build_artifact(
        self, context: Context, files: Iterable[tuple[str, Path]]
    ) -> Path:
        version: str = context.config.metadata["version"]
        dist_info = f"{context.config.metadata['name']}-{version}"

        target = context.dist_dir / f"{dist_info}.tar.gz"
        tar = tarfile.open(target, mode="w:gz", format=tarfile.PAX_FORMAT)

        try:
            for relpath, path in files:
                tar.add(
                    path,
                    arcname=os.path.join(dist_info, relpath),
                    recursive=False,
                )
                self._show_add_file(relpath, path)

            fd, temp_name = tempfile.mkstemp(prefix="pkg-info")
            pkg_info = self.format_pkginfo().encode("utf-8")
            with open(fd, "wb") as f:
                f.write(pkg_info)
            tar.add(
                temp_name, arcname=os.path.join(dist_info, "PKG-INFO"), recursive=False
            )
            self._show_add_file("PKG-INFO", Path("PKG-INFO"))
        finally:
            tar.close()

        return target



from __future__ import annotations


class BuildError(RuntimeError):
    pass


class ConfigError(ValueError):
    pass


class PDMWarning(UserWarning):
    pass


class ValidationError(ConfigError):
    def __init__(self, summary: str, details: str) -> None:
        super().__init__(summary)
        self.summary = summary
        self.details = details

    def __str__(self) -> str:
        return f"{self.summary}\n{self.details}"



import warnings
from functools import wraps
from inspect import cleandoc

from ..pre_compile import pre_compile


def _deprecated(orig, repl):
    msg = f"""
    `{orig.__module__}:{orig.__name__}` is deprecated and will be removed in future
    versions of `validate-pyproject`.

    Please use `{repl.__module__}:{repl.__name__}` instead.
    """

    @wraps(orig)
    def _wrapper(*args, **kwargs):
        warnings.warn(cleandoc(msg), category=DeprecationWarning)
        return repl(*args, **kwargs)

    return _wrapper


def vendorify(*args, **kwargs):
    return _deprecated(vendorify, pre_compile)(*args, **kwargs)
